from py2neo import Graph, Node, Relationship, NodeMatcher


class GraphUtils():

    def __init__(self, graph):
        self.graph = graph


    def get_teis_for_categories(self, categories=None, operator="AND"):
        """For given categories, this method searches database and returns TEI nodes
        that are of these categories.
        """
        if categories == None:
            categories = []
        elif not isinstance(categories, list):
            categories=[categories]

        wheres = ['"{}" IN tei.categories'.format(category) for category in categories]
        where = (" "+operator+" ").join(wheres)
        if not where:
            # if no categories are provided, make sure that at least only return
            # TEI documents that contain the categories attribute
            where = "EXISTS( tei['categories'])"

        cypher = """
            MATCH(tei :TEI)
            WHERE {}
            RETURN tei
        """
        cursor = self.graph.run(cypher.format(where))

        teis = []
        while cursor.forward():
            for tei in cursor.current:
                teis.append(tei)
        return teis


    def paragraphs_for_filename(self, filename):
        kwargs = { "filename": filename }
        cypher = """
            MATCH(paragraph :p {filename:{filename}})-[:FACS]->(zone :zone)
            RETURN paragraph
        """
        cursor = self.graph.run(
            cypher,
            parameters = {
                "filename": filename
            }
        )

        paragraphs = []
        while cursor.forward():
            for paragraph in cursor.current:
                paragraphs.append(paragraph)

        return paragraphs

    def nodes_with_link_for_filename(self, filename):
        """for a given filename, this query returns all nodes
        that contain any of the link_attributes (see below)
        The goal is to create relationships from these nodes to their targets
        """

        kwargs = { "filename": filename }
        link_attributes = ["facs", "next", "prev", "target", "ref"]
        where = " OR ".join("EXISTS(n.{})".format(attr) for attr in link_attributes)
        cypher = """
            MATCH (n)
            WHERE 
        """
        cypher += where
        cypher += " RETURN n"

        nodes = []
        while cursor.forward():
            for nodes in cursor.current:
                nodes.append(nodes)

        return nodes


    def concatenation_exists(self, node):
        cypher = """
        MATCH (node1)-[r :CONCATENATED]->(node2)
        WHERE ID(node1) = {node_id}
        RETURN r
        """
        cursor = self.graph.run(cypher, parameters={"node_id": node.identity})

        # return 1 if there exists a relation
        return cursor.forward()


    def tokens_in_paragraph(self, paragraph:Node, concatenated=0):
        """For a given paragraph, this method returns all nodes
        connected via a NEXT relationship. 
        If concatenated=1, it will return a concatenated version of the textpath.
        Returns all nodes in the found textpath.
        """

        cypher="""
        MATCH (para)-[:NEXT]->(t),
        textpath = shortestPath((t)-[:NEXT*]->(lt))
        WHERE ID(para)={paragraph_id}
        AND (para)-[:LAST]->(lt)
        AND ALL (
            rel IN relationships(textpath)
            WHERE (rel.concatenated IS NULL OR rel.concatenated = {concatenated})
        )
        RETURN nodes(textpath)
        """
        # NOTE: it is assumed that a path containing a concatenated (non-hyphened)
        # word will be always shorter than a path containing a hyphened word.
        # The non-concatenated textpath actually never has the relation-attribute 
        # rel.concatenated=0 (it is always NULL)

        cursor = self.graph.run(
            cypher, 
            parameters={
                "paragraph_id": paragraph.identity,
                "concatenated": concatenated
            }
        ) 

        nodes = [] 
        while cursor.forward():
            for entry in cursor.current:
                for node in entry:
                    nodes.append(node)
                
        return nodes


    def tokens_in_paragraphs(self,
        paragraphs,
        concatenate_words=0,
        concatenate_paragraphs=0,
        follow_metamarks=0
    ):
        """For a given list of paragraphs (nodes) this method tries to:
        - return concatenated words,
        - concatenate paragraphs, when they are connected via next and prev attributes
        - follow metamarks, when linked via target attribute
        - remove duplicate paragraphs i.e. paragraphs that have already been
          visited should not appear again.
        - this method is not recursive, i.e. search-depth is 1, which should be fine in 99% of the cases.
        """
        paragraph_visited = {}
        paragraph_dict = {}
        for paragraph in paragraphs:
            paragraph_dict[paragraph.identity] = paragraph 

        nodes = []

        for paragraph in paragraphs:
            if paragraph.identity in paragraph_visited:
                continue
            nodes.append(paragraph)
            tokens = self.tokens_in_paragraph(paragraph, concatenated=concatenate_words)
            for token in tokens:
                nodes.append(token)
                if 'metamark' in token.labels and token.get('target', False):
                    tei_file, anchor = token.get('target').split('#')
                    if anchor in paragraph_dict:
                        metamark_tokens = self.tokens_in_paragraph(
                            paragraph_dict[anchor], 
                            concatenated=concatenate_words
                        )
                        nodes += metamark_nodes
                        paragraph_visited[anchor] = True 

        return


    def create_unhyphenated(self, tokens):
        """tokens=Array of all tokens in a paragraph, as returned
        by GraphUtils.tokens_in_paragraph(para). This procedures looks for linebreaks with type=hyph
        If found, it looks forward and backwards to find the hyphened wordparts
        It then concatenates the wordparts, creates a new Node and new Releations.
        """
        tx = self.graph.begin()

        for i, token in enumerate(tokens):
            j=0
            k=0
            if token.has_label('lb'):
                wordstart = None
                wordend = None
                if token.get('type') == 'hyph':
                    
                    # walk back and find a token which is a wordpart
                    # and not any punctuation or similar
                    j=i-1
                    while j>0:
                        #print("j = {}".format(j))
                        if tokens[j].has_label('token') \
                        and tokens[j]["string"] \
                        and not tokens[j]["is_punct"]: 
                            #print("TRY: "+str(tokens[j]))
                            wordstart = tokens[j]
                            break
                        else:
                            j = j-1

                    # walk forward and find a token
                    k=i+1
                    while k>0 and tokens[k]:
                        #print("k = {}".format(k))
                        if tokens[k].has_label('token'):
                            #print("TRY END: "+str(tokens[k]))
                            wordend = tokens[k]
                            break
                        else:
                            k = k+1
                          
                    concat_word = ''
                    if wordstart and wordend and not self.concatenation_exists(wordstart):
                        #print("---START: "+str(wordstart))
                        #print("---  END: "+str(wordend))
                        if any(
                            wordstart["string"].endswith(s) for s\
                            in ['-', '\N{NOT SIGN}', '\N{NON-BREAKING HYPHEN}']
                        ):
                            concat_word = wordstart["string"][:-1]
                        else:
                            concat_word = wordstart["string"]
                            
                        concat_word += wordend["string"]
                        
                        # create new concatenated token
                        # with blank as whitespace
                        labels = list(
                            set(
                                ['token', 'concatenated'] + list(wordstart.labels) + list(wordend.labels)
                            )
                        )
                        attrs = { 
                            "string"    : concat_word,
                            "whitespace": wordend["whitespace"],
                            "filename"  : wordstart["filename"],
                            "idno"      : wordstart["idno"],
                        }
                        concat_node = Node(
                            *labels,
                            **attrs
                        )
                        #print("+ {}".format(concat_word))
                        tx.create(concat_node)
                        
                        # create relations from hyphened wordpards to concatenated word
                        rs = Relationship(
                            wordstart,
                            "CONCATENATED",
                            concat_node
                        )
                        rs2 = Relationship(
                            wordend,
                            "CONCATENATED",
                            concat_node
                        )
                        tx.create(rs)                
                        tx.create(rs2)
                        
                        # create direct connection for non-hyphened version
                        # of the thext
                        if j >0:
                            before_wordstart = tokens[j-1]
                            rs3 = Relationship(
                                before_wordstart,
                                "NEXT",
                                concat_node,
                                concatenated=1
                            )
                            tx.create(rs3)
                            
                        if len(tokens) > k+1:
                            after_wordend = tokens[k+1]
                            rs4 = Relationship(
                                concat_node,
                                "NEXT",
                                after_wordend,
                                concatenated=1
                            )
                            tx.create(rs4)

        tx.commit()


    def link_inner_relationships(self, filename):
        """creates new relationships between nodes of a given filename,
        that contain one of the link attributes below which point to
        a node inside the same xml file. Ignores all :join nodes, because
        they are handled by the handle_join_elements method.

        Within a link attribute, pointers can be separated by whitespace
        https://www.tei-c.org/release/doc/tei-p5-doc/en/html/ref-att.global.linking.html

        This method should be only run once after a TEI document has
        been parsed.
        Does not return anything.
        """
        
        link_attributes = ["facs", "next", "prev", "target", "ref"]

        cypher = """
	MATCH (from_node)
	WHERE from_node.filename = {filename}
        AND NOT from_node:join
        AND from_node[{link_attribute}] IS NOT NULL
	WITH from_node, SPLIT(from_node[{link_attribute}], " ") AS pointers
	UNWIND pointers as pointer
	WITH from_node, SPLIT(pointer, "#") AS target
        WITH from_node, target, CASE WHEN target[0] = "" THEN {filename} ELSE target[0] END AS filename
	MERGE (to_node {filename: {filename}, `xml:id`:target[1]} )
        """

        for link_attribute in link_attributes:
            parameters = {
                "filename"      : filename,
                "link_attribute": link_attribute,
            }
            create_rel = " CREATE (from_node)-[r:{}]->(to_node)".format(
                link_attribute.upper()
            )
            cursor = self.graph.run(cypher+create_rel, parameters = parameters)


    def handle_join_elements(self, filename):
        """Searches for all <join> elements for a given filename.
        Creates relationships between the mentioned elements of the «target» attribute
        removes the <join> elements after the relations have been established.
        Does not return anything.
        """
        
        cypher = """
            MATCH (join_node :join)
            WHERE join_node.filename = {filename}
            AND join_node['target'] IS NOT NULL
            WITH join_node, SPLIT(join_node['target'], " ") AS endpoints
            WITH join_node, SPLIT(endpoints[0], "#") AS source, SPLIT(endpoints[1], "#") AS target
            WITH join_node, source, target,
            CASE WHEN source[0] = "" THEN {filename} ELSE source[0] END AS source_filename,
            CASE WHEN target[0] = "" THEN {filename} ELSE target[0] END AS target_filename
            MERGE (from_node {filename: source_filename, `xml:id`:source[1]})
            MERGE (to_node {filename: target_filename, `xml:id`:target[1]})
            MERGE (from_node)-[r :JOIN]->(to_node)
            DETACH DELETE join_node
        """

        parameters = { "filename": filename }
        cursor = self.graph.run(cypher, parameters = parameters)

